@@ -8,11 +8,12 @@ from typing import List, Optional, Type, Union
 from django.contrib.gis.db import models as gis_models
 from django.contrib.gis.db.models import Extent, GeometryField
 from django.contrib.gis.db.models.functions import Transform
-from django.core.exceptions import ImproperlyConfigured
+from django.core.exceptions import FieldDoesNotExist, ImproperlyConfigured
 from django.db import models
 from django.db.models.fields.reverse_related import ForeignObjectRel
 from django.utils.functional import cached_property  # py3.8: functools
 
+from gisserver.exceptions import ExternalValueError
 from gisserver.types import (
     XPathMatch,
     XsdAnyType,
@@ -35,6 +36,7 @@ __all__ = [
     "field",
     "FeatureField",
     "ComplexFeatureField",
+    "get_basic_field_type",
 ]
 
 XSD_TYPES = {
@@ -62,7 +64,7 @@ XSD_TYPES = {
 DEFAULT_XSD_TYPE = XsdTypes.anyType
 
 
-def _get_basic_field_type(field_name: str, model_field: models.Field) -> XsdAnyType:
+def get_basic_field_type(field_name: str, model_field: models.Field) -> XsdAnyType:
     """Determine the XSD field type for a Django field."""
     try:
         # Direct instance, quickly resolved!
@@ -72,10 +74,10 @@ def _get_basic_field_type(field_name: str, model_field: models.Field) -> XsdAnyT
 
     if isinstance(model_field, models.ForeignKey):
         # Don't let it query on the relation value yet
-        return _get_basic_field_type(field_name, model_field.target_field)
+        return get_basic_field_type(field_name, model_field.target_field)
     elif isinstance(model_field, ForeignObjectRel):
         # e.g. ManyToOneRel descriptor of a foreignkey_id field.
-        return _get_basic_field_type(field_name, model_field.remote_field.target_field)
+        return get_basic_field_type(field_name, model_field.remote_field.target_field)
     else:
         # Subclass checks:
         for field_cls, xsd_type in XSD_TYPES.items():
@@ -90,13 +92,28 @@ def _get_basic_field_type(field_name: str, model_field: models.Field) -> XsdAnyT
         return DEFAULT_XSD_TYPE
 
 
-def _get_model_fields(model, fields):
+def _get_target_model(model_field) -> Type[models.Model]:
+    """Find which model a related field points at"""
+    if isinstance(model_field, models.ForeignKey):
+        return model_field.target_field.model
+    elif isinstance(model_field, ForeignObjectRel):
+        # e.g. ManyToOneRel descriptor of a foreignkey_id field.
+        return model_field.remote_field.model
+    else:
+        raise NotImplementedError(
+            f"Model field is not supported as relation: {model_field.__class__.__name__}"
+        )
+
+
+def _get_model_fields(model, fields, parent=None):
     if fields == "__all__":
         # All regular fields
         return [
             FeatureField(
                 name=f.attname if isinstance(f, models.ForeignKey) else f.name,
+                # .bind() is called directly:
                 model=model,
+                parent=parent,
             )
             for f in model._meta.get_fields()
         ]
@@ -104,7 +121,7 @@ def _get_model_fields(model, fields):
         # Only defined fields
         fields = [f if isinstance(f, FeatureField) else FeatureField(f) for f in fields]
         for field in fields:
-            field.bind(model)
+            field.bind(model, parent=parent)
         return fields
 
 
@@ -128,13 +145,25 @@ class FeatureField:
     an XSD definition that the remaining application uses.
     """
 
+    #: The class that's used to generate the XsdElement.
+    xsd_element_class: Type[XsdElement] = XsdElement
+
     model: Optional[Type[models.Model]]
     model_field: Optional[models.Field]
 
-    def __init__(self, name, model=None):
+    def __init__(
+        self,
+        name,
+        model_attribute=None,
+        model=None,
+        parent: "Optional[ComplexFeatureField]" = None,
+    ):
         self.name = name
+        self.model_attribute = model_attribute
         self.model = None
         self.model_field = None
+        self.parent = parent
+        self._nillable_relation = False
         if model is not None:
             self.bind(model)
 
@@ -142,26 +171,64 @@ class FeatureField:
         return f"<{self.__class__.__name__}: {self.name}>"
 
     def _get_xsd_type(self):
-        return _get_basic_field_type(self.name, self.model_field)
+        return get_basic_field_type(self.name, self.model_field)
 
-    def bind(self, model: Type[models.Model]):
-        """Late-binding for the model"""
+    def bind(
+        self, model: Type[models.Model], parent: "Optional[ComplexFeatureField]" = None,
+    ):
+        """Late-binding for the model.
+
+        This method is called internally when the field definition wasn't
+        linked to a model yet. This allows the fields to be defined first,
+        in external code, and then become part of the ``FeatureType`` fields
+        list.
+
+        :param model: The model is field is linked to.
+        :param parent: When this element is part of a complex feature,
+                       this links to the parent field.
+        """
         if self.model is not None:
             raise RuntimeError(f"Feature field '{self.name}' cannot be reused")
         self.model = model
-        self.model_field = self.model._meta.get_field(self.name)
+        self.parent = parent
+
+        if self.model_attribute:
+            # Support dot based field traversal.
+            field_path = self.model_attribute.split(".")
+            try:
+                field: models.Field = self.model._meta.get_field(field_path[0])
+            except FieldDoesNotExist as e:
+                raise ImproperlyConfigured(
+                    f"FeatureField '{self.name}' has an invalid"
+                    f" model_attribute: '{self.model_attribute}' can't be"
+                    f" resolved for model '{self.model.__name__}'."
+                ) from e
+            for name in field_path[1:]:
+                if field.null:
+                    self._nillable_relation = True
+                field = _get_target_model(field)._meta.get_field(name)
+            self.model_field = field
+        else:
+            self.model_field = self.model._meta.get_field(self.name)
 
     @cached_property
     def xsd_element(self) -> XsdElement:
         """Define the XMLSchema definition for a model field.
-        This is used in DescribeFeatureType.
+
+        This definition is used by the remaining application to access the
+        data. It's the basis for DescribeFeatureType, and it's ``get_value()``
+        method is read to access the model field data.
         """
-        return XsdElement(
+        if self.model_field is None:
+            raise RuntimeError(f"bind() was not called for {self!r}")
+
+        return self.xsd_element_class(
             name=self.name,
             type=self._get_xsd_type(),
-            nillable=self.model_field.null,
+            nillable=self.model_field.null or self._nillable_relation,
             min_occurs=0,
             max_occurs=1 if isinstance(self.model_field, GeometryField) else None,
+            model_attribute=self.model_attribute,
             source=self.model_field,
         )
 
@@ -176,19 +243,21 @@ class ComplexFeatureField(FeatureField):
     This translates the foreign key into an embedded XSD complex type.
     """
 
-    def __init__(self, name: str, fields: _FieldDefinitions):
+    def __init__(
+        self, name: str, fields: _FieldDefinitions, model_attribute=None, model=None
+    ):
         """
         :param name: Name of the model field.
         :param fields: List of fields to expose for the target model. This can
             be a list of :class:`FeatureField` objects, or plain field names.
             Using ``__all__`` also works but is not recommended outside testing.
         """
-        super().__init__(name)
+        super().__init__(name, model_attribute=model_attribute, model=model)
         self._fields = fields
 
     def _get_xsd_type(self) -> XsdComplexType:
         """Generate the XSD description for the field with an object relation."""
-        fields = _get_model_fields(self.target_model, self._fields)
+        fields = _get_model_fields(self.target_model, self._fields, parent=self)
         return XsdComplexType(
             name=f"{self.target_model._meta.object_name}Type",
             elements=[field.xsd_element for field in fields],
@@ -201,19 +270,18 @@ class ComplexFeatureField(FeatureField):
         if self.model_field is None:
             raise RuntimeError("FeatureField.bind() is not called yet")
 
-        if isinstance(self.model_field, models.ForeignKey):
-            return self.model_field.target_field.model
-        elif isinstance(self.model_field, ForeignObjectRel):
-            # e.g. ManyToOneRel descriptor of a foreignkey_id field.
-            return self.model_field.remote_field.model
-        else:
+        try:
+            return _get_target_model(self.model_field)
+        except NotImplementedError:
             raise ImproperlyConfigured(
                 f"{self.__class__.__name__} does not support fields of "
                 f"type {self.model_field.__class__.__name__}."
-            )
+            ) from None
 
 
-def field(name: str, fields: Optional[_FieldDefinitions] = None) -> FeatureField:
+def field(
+    name: str, *, model_attribute=None, fields: Optional[_FieldDefinitions] = None
+) -> FeatureField:
     """Shortcut to define a WFS field.
 
     This automatically selects the proper field class,
@@ -225,9 +293,11 @@ def field(name: str, fields: Optional[_FieldDefinitions] = None) -> FeatureField
         Using ``__all__`` also works but is not recommended outside testing.
     """
     if fields is not None:
-        return ComplexFeatureField(name=name, fields=fields)
+        return ComplexFeatureField(
+            name=name, model_attribute=model_attribute, fields=fields
+        )
     else:
-        return FeatureField(name)
+        return FeatureField(name, model_attribute=model_attribute)
 
 
 class FeatureType:
@@ -289,7 +359,7 @@ class FeatureType:
         self.metadata_url = metadata_url
 
         # Validate that the name doesn't require XML escaping.
-        if html.escape(self.name) != self.name or " " in self.name:
+        if html.escape(self.name) != self.name or " " in self.name or ":" in self.name:
             raise ValueError(f"Invalid feature name for XML: <app:{self.name}>")
 
         # Auto-detect geometry fields (also fills geometry_field_name)
@@ -306,6 +376,11 @@ class FeatureType:
         """
         pass
 
+    @cached_property
+    def xml_name(self):
+        """Return the feature name with xml namespace prefix."""
+        return f"app:{self.name}"
+
     @cached_property
     def geometry_field_names(self):
         return {f.name for f in self.geometry_fields}
@@ -315,7 +390,7 @@ class FeatureType:
         """Define which fields to render."""
         # This lazy reading allows providing 'fields' as lazy value.
         if self._fields is None:
-            return [FeatureField(self.geometry_field_name, self.model)]
+            return [FeatureField(self.geometry_field_name, model=self.model)]
         else:
             return _get_model_fields(self.model, self._fields)
 
@@ -410,7 +485,7 @@ class FeatureType:
         """
         path = self._cached_resolver(xpath)  # calls _inner_resolve_element
         if path is None:
-            raise ValueError(f"Field '{xpath}' does not exist.")
+            raise ExternalValueError(f"Field '{xpath}' does not exist.")
 
         return XPathMatch(path, query=xpath)
 
