@@ -8,7 +8,7 @@ import re
 from datetime import date, datetime, time
 from functools import reduce
 from html import escape
-from typing import Optional, cast
+from typing import List, Optional, cast
 
 from django.contrib.gis import geos
 from django.contrib.gis.db.models.functions import AsGML, Transform, Union
@@ -26,6 +26,7 @@ from .base import (
     OutputRenderer,
     StringBuffer,
     build_db_annotations,
+    get_db_annotation,
     get_db_geometry_selects,
     get_db_geometry_target,
 )
@@ -108,21 +109,23 @@ class GML32Renderer(OutputRenderer):
             "http://www.opengis.net/gml/3.2 http://schemas.opengis.net/gml/3.2.1/gml.xsd",
         ]
 
-        return """
-     xmlns:app="{app_xml_namespace}"
-     xmlns:gml="http://www.opengis.net/gml/3.2"
-     xmlns:wfs="http://www.opengis.net/wfs/2.0"
-     xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance"
-     xsi:schemaLocation="{schema_location}"
-""".format(
+        return (
+            'xmlns:wfs="http://www.opengis.net/wfs/2.0" '
+            'xmlns:gml="http://www.opengis.net/gml/3.2" '
+            'xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance" '
+            'xmlns:app="{app_xml_namespace}" '
+            'xsi:schemaLocation="{schema_location}"'
+        ).format(
             app_xml_namespace=escape(self.app_xml_namespace),
             schema_location=escape(" ".join(schema_location)),
         )
 
     def render_xmlns_standalone(self):
         """Generate the xmlns block that the documente needs"""
+        # xsi is needed for "xsi:nil="true"' attributes.
         return (
             f' xmlns:app="{escape(self.app_xml_namespace)}"'
+            ' xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance"'
             ' xmlns:gml="http://www.opengis.net/gml/3.2"'
         )
 
@@ -134,7 +137,7 @@ class GML32Renderer(OutputRenderer):
             raise NotFound("Feature not found.")
 
         output = StringBuffer()
-        output.write("""<?xml version='1.0' encoding="UTF-8" ?>\n""")
+        output.write('<?xml version="1.0" encoding="UTF-8"?>\n')
         output.write(
             self.render_wfs_member(
                 feature_type=sub_collection.feature_type,
@@ -164,11 +167,11 @@ class GML32Renderer(OutputRenderer):
 
         output.write(
             f"""<?xml version='1.0' encoding="UTF-8" ?>\n"""
-            f"<wfs:{self.xml_collection_tag}\n"
-            f"     {xmlns}\n"
-            f'     timeStamp="{collection.timestamp}"'
+            f"<wfs:{self.xml_collection_tag} {xmlns}"
+            f' timeStamp="{collection.timestamp}"'
             f' numberMatched="{number_matched}"'
-            f' numberReturned="{int(number_returned)}"{next}{previous}>\n'
+            f' numberReturned="{int(number_returned)}"'
+            f"{next}{previous}>\n"
         )
 
         if number_returned:
@@ -401,7 +404,13 @@ class DBGML32Renderer(GML32Renderer):
     """Faster GetFeature renderer that uses the database to render GML 3.2"""
 
     @classmethod
-    def decorate_queryset(cls, feature_type, queryset, output_crs, **params):
+    def decorate_queryset(
+        cls,
+        feature_type: FeatureType,
+        queryset: models.QuerySet,
+        output_crs: CRS,
+        **params,
+    ):
         """Update the queryset to let the database render the GML output.
         This is far more efficient then GeoDjango's logic, which performs a
         C-API call for every single coordinate of a geometry.
@@ -410,7 +419,9 @@ class DBGML32Renderer(GML32Renderer):
             feature_type, queryset, output_crs, **params
         )
 
-        geometries = get_db_geometry_selects(feature_type.xsd_type, output_crs)
+        geometries = get_db_geometry_selects(
+            feature_type.xsd_type.gml_elements, output_crs
+        )
         return queryset.defer(*geometries.keys()).annotate(
             _as_envelope_gml=cls.get_db_envelope_as_gml(
                 feature_type, queryset, output_crs
@@ -419,14 +430,33 @@ class DBGML32Renderer(GML32Renderer):
         )
 
     @classmethod
-    def get_prefetch_queryset(cls, xsd_element: XsdElement, output_crs: CRS):
-        """Perform DB annotations for the prefetched relation too."""
-        xsd_type: XsdComplexType = cast(XsdComplexType, xsd_element.type)
-        model = xsd_type.source
+    def get_prefetch_queryset(
+        cls,
+        feature_type: FeatureType,
+        xsd_elements: List[XsdElement],
+        fields: List[str],
+        output_crs: CRS,
+    ):
+        """Perform DB annotations for prefetched relations too."""
+        model = xsd_elements[0].source
+        if model is None:
+            return None
 
-        geometries = get_db_geometry_selects(xsd_type, output_crs)
+        # Find which fields are GML elements
+        gml_elements = []
+        for e in xsd_elements:
+            if e.is_gml:
+                # Prefetching a flattened relation
+                gml_elements.append(e)
+            elif e.type.is_complex_type:
+                # Prefetching a complex type
+                xsd_type: XsdComplexType = cast(XsdComplexType, e.type)
+                gml_elements.extend(xsd_type.gml_elements)
+
+        geometries = get_db_geometry_selects(gml_elements, output_crs)
+        fields = set(fields) - set(geometries.keys())
         if geometries:
-            return model.objects.defer(*geometries.keys()).annotate(
+            return model.objects.only("pk", *fields).annotate(
                 **build_db_annotations(geometries, "_as_gml_{name}", _AsGML),
             )
 
@@ -464,7 +494,7 @@ class DBGML32Renderer(GML32Renderer):
     ):
         if xsd_element.is_gml:
             # Optimized path, pre-rendered GML
-            value = getattr(instance, f"_as_gml_{xsd_element.name}")
+            value = get_db_annotation(instance, xsd_element.name, "_as_gml_{name}")
             if value is None:
                 # Avoid incrementing gml_seq
                 return f'      <app:{xsd_element.name} xsi:nil="true" />\n'
@@ -514,6 +544,17 @@ class GML32ValueRenderer(GML32Renderer):
         self.value_reference = value_reference
         super().__init__(*args, **kwargs)
 
+    @classmethod
+    def decorate_queryset(
+        cls,
+        feature_type: FeatureType,
+        queryset: models.QuerySet,
+        output_crs: CRS,
+        **params,
+    ):
+        # Don't optimize queryset, it only retrieves one value
+        return queryset
+
     def render_wfs_member(
         self, feature_type: FeatureType, instance: dict, extra_xmlns=""
     ) -> str:
