@@ -3,11 +3,13 @@ from .filters import Filter
 from .functions import function_registry
 from .identifiers import ResourceId
 from .query import CompiledQuery
+from .sorting import SortBy
 
 __all__ = [
     "Filter",
     "CompiledQuery",
     "ValueReference",
     "ResourceId",
+    "SortBy",
     "function_registry",
 ]
