@@ -4,7 +4,7 @@ The class names and attributes are identical to those in the FES spec.
 from itertools import groupby
 
 import operator
-from dataclasses import dataclass
+from dataclasses import dataclass, field
 from decimal import Decimal
 from enum import Enum
 from functools import reduce
@@ -14,6 +14,7 @@ from xml.etree.ElementTree import Element
 from django.contrib.gis import measure
 from django.db.models import Q
 
+from gisserver.exceptions import ExternalValueError
 from gisserver.parsers import gml
 from gisserver.parsers.base import FES20, BaseNode, TagNameEnum, tag_registry
 from gisserver.parsers.utils import expect_tag, get_attribute, get_child
@@ -171,13 +172,35 @@ class IdOperator(Operator):
 class NonIdOperator(Operator):
     """Abstract base class, as defined by FES spec."""
 
+    _source = None
+
     def build_compare(
-        self, compiler: CompiledQuery, lhs: Expression, lookup: str, rhs: RhsTypes
+        self,
+        compiler: CompiledQuery,
+        lhs: Expression,
+        lookup: str,
+        rhs: Union[Expression, RhsTypes],
     ) -> Q:
         """Use the value in comparison with some other expression.
 
         This calls build_lhs() and build_rhs() on the expressions.
         """
+        # lhs and rhs are allowed to be reversed. However, the SQL compiler
+        # works much simpler when Django can predict the actual data type.
+        if isinstance(rhs, Literal) and isinstance(rhs, ValueReference):
+            lhs, rhs = rhs, lhs
+
+        if (
+            compiler.feature_type is not None
+            and isinstance(lhs, ValueReference)
+            and isinstance(rhs, Literal)
+        ):
+            # When a common case of value comparison is done, the inputs
+            # can be validated before the ORM query is constructed.
+            xsd_element = compiler.feature_type.resolve_element(lhs.xpath).child
+            tag = self._source if self._source is not None else None
+            xsd_element.validate_comparison(rhs.raw_value, lookup=lookup, tag=tag)
+
         lhs = lhs.build_lhs(compiler)
 
         if isinstance(rhs, (Expression, gml.GM_Object)):
@@ -219,14 +242,15 @@ class DistanceOperator(SpatialOperator):
     operatorType: DistanceOperatorName
     geometry: gml.GM_Object
     distance: Measure
+    _source: Optional[str] = field(compare=False, default=None)
 
     @classmethod
     def from_xml(cls, element: Element):
         geometries = gml.find_gml_nodes(element)
         if not geometries:
-            raise ValueError(f"Missing gml element in <{element.tag}>")
+            raise ExternalValueError(f"Missing gml element in <{element.tag}>")
         elif len(geometries) > 1:
-            raise ValueError(f"Multiple gml elements found in <{element.tag}>")
+            raise ExternalValueError(f"Multiple gml elements found in <{element.tag}>")
 
         return cls(
             valueReference=ValueReference.from_xml(
@@ -235,6 +259,7 @@ class DistanceOperator(SpatialOperator):
             operatorType=DistanceOperatorName.from_xml(element),
             geometry=gml.parse_gml_node(geometries[0]),
             distance=Measure.from_xml(get_child(element, FES20, "Distance")),
+            _source=element.tag,
         )
 
     def build_query(self, compiler: CompiledQuery) -> Q:
@@ -254,27 +279,36 @@ class BinarySpatialOperator(SpatialOperator):
     operatorType: SpatialOperatorName
     operand1: Optional[ValueReference]
     operand2: SpatialDescription
+    _source: Optional[str] = field(compare=False, default=None)
 
     @classmethod
     def from_xml(cls, element: Element):
+        operator_type = SpatialOperatorName.from_xml(element)
+        if operator_type is SpatialOperatorName.BBOX and len(element) == 1:
+            # For BBOX, the geometry operator is optional
+            ref = None
+            geo = element[0]
+        else:
+            if len(element) != 2:
+                raise ExternalValueError(f"{element.tag} should have 2 operators")
+            ref, geo = list(element)
+
         return cls(
-            operatorType=SpatialOperatorName.from_xml(element),
-            operand1=ValueReference.from_xml(element[0]),
+            operatorType=operator_type,
+            operand1=ValueReference.from_xml(ref) if ref is not None else None,
             operand2=tag_registry.from_child_xml(
-                element[1],
-                allowed_types=SpatialDescription.__args__,  # get_args() in 3.8
+                geo, allowed_types=SpatialDescription.__args__,  # get_args() in 3.8
             ),
+            _source=element.tag,
         )
 
     def build_query(self, compiler: CompiledQuery) -> Q:
-        if self.operand1 is None:
-            raise NotImplementedError()
+        operant1 = self.operand1
+        if operant1 is None:
+            operant1 = ValueReference(xpath=compiler.feature_type.geometry_field_name)
 
         return self.build_compare(
-            compiler,
-            lhs=self.operand1,
-            lookup=self.operatorType.value,
-            rhs=self.operand2,
+            compiler, lhs=operant1, lookup=self.operatorType.value, rhs=self.operand2,
         )
 
 
@@ -286,6 +320,7 @@ class TemporalOperator(NonIdOperator):
     operatorType: TemporalOperatorName
     operand1: ValueReference
     operand2: TemporalOperand
+    _source: Optional[str] = field(compare=False, default=None)
 
     @classmethod
     def from_xml(cls, element: Element):
@@ -295,6 +330,7 @@ class TemporalOperator(NonIdOperator):
             operand2=tag_registry.from_child_xml(
                 element[1], allowed_types=TemporalOperand.__args__,  # get_args() in 3.8
             ),
+            _source=element.tag,
         )
 
 
@@ -315,6 +351,7 @@ class BinaryComparisonOperator(ComparisonOperator):
     expression: Tuple[Expression, Expression]
     matchCase: bool = True
     matchAction: MatchAction = MatchAction.Any
+    _source: Optional[str] = field(compare=False, default=None)
 
     @classmethod
     def from_xml(cls, element: Element):
@@ -328,6 +365,7 @@ class BinaryComparisonOperator(ComparisonOperator):
             matchAction=MatchAction(
                 element.get("matchAction", default=MatchAction.Any)
             ),
+            _source=element.tag,
         )
 
     def build_query(self, compiler: CompiledQuery) -> Q:
@@ -345,6 +383,7 @@ class BetweenComparisonOperator(ComparisonOperator):
     expression: Expression
     lowerBoundary: Expression
     upperBoundary: Expression
+    _source: Optional[str] = field(compare=False, default=None)
 
     @classmethod
     def from_xml(cls, element: Element):
@@ -354,6 +393,7 @@ class BetweenComparisonOperator(ComparisonOperator):
             expression=Expression.from_child_xml(element[0]),
             lowerBoundary=Expression.from_child_xml(lower[0]),
             upperBoundary=Expression.from_child_xml(upper[0]),
+            _source=element.tag,
         )
 
     def build_query(self, compiler: CompiledQuery) -> Q:
@@ -374,6 +414,7 @@ class LikeOperator(ComparisonOperator):
     wildCard: str
     singleChar: str
     escapeChar: str
+    _source: Optional[str] = field(compare=False, default=None)
 
     @classmethod
     def from_xml(cls, element: Element):
@@ -385,12 +426,14 @@ class LikeOperator(ComparisonOperator):
             wildCard=get_attribute(element, "wildCard"),
             singleChar=get_attribute(element, "singleChar"),
             escapeChar=get_attribute(element, "escapeChar"),
+            _source=element.tag,
         )
 
     def build_query(self, compiler: CompiledQuery) -> Q:
         lhs, rhs = self.expression
         if isinstance(rhs, Literal):
-            value = rhs.value
+            value = str(rhs.value)  # value could be auto-casted to int.
+
             # Not using r"\" here as that is a syntax error.
             if self.escapeChar != "\\":
                 value = value.replace("\\", "\\\\").replace(self.escapeChar, "\\")
@@ -399,7 +442,7 @@ class LikeOperator(ComparisonOperator):
             if self.singleChar != "_":
                 value = value.replace("_", r"\_").replace(self.singleChar, "_")
 
-            rhs = value
+            rhs = Literal(raw_value=value)
         else:
             raise NotImplementedError()
 
@@ -416,12 +459,14 @@ class NilOperator(ComparisonOperator):
 
     expression: Optional[Expression]
     nilReason: str
+    _source: Optional[str] = field(compare=False, default=None)
 
     @classmethod
     def from_xml(cls, element: Element):
         return cls(
             expression=Expression.from_child_xml(element[0]) if element else None,
             nilReason=element.get("nilReason"),
+            _source=element.tag,
         )
 
     def build_query(self, compiler: CompiledQuery) -> Q:
@@ -439,10 +484,13 @@ class NullOperator(ComparisonOperator):
     """
 
     expression: Expression
+    _source: Optional[str] = field(compare=False, default=None)
 
     @classmethod
     def from_xml(cls, element: Element):
-        return cls(expression=Expression.from_child_xml(element[0]))
+        return cls(
+            expression=Expression.from_child_xml(element[0]), _source=element.tag
+        )
 
     def build_query(self, compiler: CompiledQuery) -> Q:
         # For now, the implementation is identical to PropertyIsNil.
@@ -465,12 +513,14 @@ class BinaryLogicOperator(LogicalOperator):
 
     operands: List[NonIdOperator]
     operatorType: BinaryLogicType
+    _source: Optional[str] = field(compare=False, default=None)
 
     @classmethod
     def from_xml(cls, element: Element):
         return cls(
             operands=[NonIdOperator.from_child_xml(child) for child in element],
             operatorType=BinaryLogicType.from_xml(element),
+            _source=element.tag,
         )
 
     def build_query(self, compiler: CompiledQuery) -> Q:
@@ -486,12 +536,14 @@ class UnaryLogicOperator(LogicalOperator):
 
     operands: NonIdOperator
     operatorType: UnaryLogicType
+    _source: Optional[str] = field(compare=False, default=None)
 
     @classmethod
     def from_xml(cls, element: Element):
         return cls(
             operands=NonIdOperator.from_child_xml(element[0]),
             operatorType=UnaryLogicType.from_xml(element),
+            _source=element.tag,
         )
 
     def build_query(self, compiler: CompiledQuery) -> Q:
